@@ -5,15 +5,29 @@ main application class for GUIApp-conform Kivy app
 This ae portion is providing two classes (:class:`FrameworkApp`
 and :class:`KivyMainApp` and some useful constants.
 
-The class :class:`KivyMainApp` is implementing a abstract main app
-class, based on the abstract base class :class:`~ae.gui_app.MainAppBase`
-that is reducing the amount of code needed for to create a Python
-application based on the :ref:`kivy <kivy.org>` framework.
+The class :class:`KivyMainApp` is implementing a main app
+class that is reducing the amount of code needed for
+to create a Python application based on the
+:ref:`kivy framework<kivy.org>`.
+
+:class:`KivyMainApp` is based on the following classes:
+
+* the abstract base class :class:`~ae.gui_app.MainAppBase`
+  which adds the concepts of :ref:`application status`
+  (including :ref:`app-state-variables` and :ref:`app-state-constants`),
+  :ref:`application flow` and :ref:`application events`.
+* the class :class:`~ae.console.ConsoleApp` is adding
+  :ref:`config-files`, :ref:`config-variables`
+  and :ref:`config-options`.
+* the class :class:`~ae.core.AppBase` is adding
+  :ref:`application logging` and :ref:`application debugging`.
+
 
 The main app class :class:`KivyMainApp` is also encapsulating the
-:class:`Kivy app class <kivy.app.App>` within the :class:`FrameworkApp`. The
-instance of the Kivy app class can be directly accessed from the main app class
-instance via the :attr:`~KivyMainApp.framework_app` attribute.
+:class:`Kivy app class <kivy.app.App>` within the :class:`FrameworkApp`.
+An instance of the Kivy app class can be directly accessed from
+the main app class instance via the
+:attr:`~KivyMainApp.framework_app` attribute.
 
 
 unit tests
@@ -41,8 +55,10 @@ from kivy.core.window import Window                                         # ty
 from kivy.factory import Factory, FactoryException                          # type: ignore
 from kivy.lang import Builder                                               # type: ignore
 # pylint: disable=no-name-in-module
-from kivy.properties import BooleanProperty, DictProperty, ListProperty     # type: ignore
+from kivy.properties import BooleanProperty, DictProperty, ListProperty, ObjectProperty  # type: ignore
 from kivy.core.audio import SoundLoader                                     # type: ignore
+from kivy.uix.dropdown import DropDown                                      # type: ignore
+from kivy.uix.popup import Popup                                            # type: ignore
 from kivy.uix.widget import Widget                                          # type: ignore
 from plyer import vibrator                                                  # type: ignore
 
@@ -55,13 +71,18 @@ from ae.gui_app import (                                                    # ty
 )                                                                           # type: ignore
 
 
-__version__ = '0.0.17'
+__version__ = '0.0.24'
 
 
 kivy.require('1.9.1')  # currently using 1.11.1 but at least 1.9.1 is needed for Window.softinput_mode 'below_target'
-# Window.softinput_mode = 'below_target'  # ensure android keyboard is not covering Popup/text input
 
-MAIN_KV_FILE_NAME = 'main.kv'       #: default file name of the main kv file
+# if the entry field is on top of the screen then it will be disappear with below_target mode
+# and in the default mode ('') the keyboard will cover the entry field if it is in the lower part of the screen
+# therefore commented out the following two code lines (and setting it now depending on the entry field y position)
+# if Window:                                  # is None on gitlab ci
+#    Window.softinput_mode = 'below_target'  # ensure android keyboard is not covering Popup/text input if at bottom
+
+MAIN_KV_FILE_NAME = 'main.kv'           #: default file name of the main kv file
 
 LOVE_VIBRATE_PATTERN = (0.0, 0.12, 0.12, 0.21, 0.03, 0.12, 0.12, 0.12)
 """ short/~1.2s vibrate pattern for fun/love notification. """
@@ -78,6 +99,10 @@ CRITICAL_VIBRATE_PATTERN = (0.00, 0.12, 0.12, 0.12, 0.12, 0.12,
 WIDGETS = '''\
 #: import Window kivy.core.window.Window
 
+#: import DEBUG_LEVELS ae.core.DEBUG_LEVELS
+#: import DEF_LANGUAGE ae.i18n.DEF_LANGUAGE
+#: import installed_languages ae.i18n.installed_languages
+
 #: import MIN_FONT_SIZE ae.gui_app.MIN_FONT_SIZE
 #: import MAX_FONT_SIZE ae.gui_app.MAX_FONT_SIZE
 #: import THEME_LIGHT_BACKGROUND_COLOR ae.gui_app.THEME_LIGHT_BACKGROUND_COLOR
@@ -89,6 +114,7 @@ WIDGETS = '''\
 #: import flow_key ae.gui_app.flow_key
 #: import flow_key_split ae.gui_app.flow_key_split
 
+
 <ThemeButton@ButtonBehavior+Label>:
     circle_fill_color: 0, 0, 0, 0
     square_fill_color: 0, 0, 0, 0
@@ -127,6 +153,7 @@ WIDGETS = '''\
     foreground_color: app.font_color
     background_color: Window.clearcolor
 
+
 <FlowButton@ThemeButton>:
     ae_flow_id: ''
     ae_clicked_kwargs: dict(popup_kwargs=dict(parent=self))
@@ -140,14 +167,14 @@ WIDGETS = '''\
 <OptionalButton@FlowButton>:
     visible: True
     size_hint: None, None
-    width: self.width if self.visible else 0
     width: self.height if self.visible else 0
+    height: self.height if self.visible else 0
     disabled: not self.visible
     opacity: 1 if self.visible else 0
 
 
 # DropDown flow gets handled identical like for a Popup
-<FlowDropDown@DropDown>:
+<FlowDropDown>:
     ae_closed_kwargs: dict(flow_id=id_of_flow('', '')) if app.main_app.flow_path_action(-2) in ('', 'enter') else dict()
     on_dismiss: app.main_app.change_flow(id_of_flow('close', 'flow_popup'), **self.ae_closed_kwargs)
     auto_width: False
@@ -160,7 +187,7 @@ WIDGETS = '''\
             rounded_rectangle: self.x, self.y, self.width, self.height, sp(9)
 
 
-<FlowPopup@Popup>:
+<FlowPopup>:
     ae_closed_kwargs: dict(flow_id=id_of_flow('', '')) if app.main_app.flow_path_action(-2) in ('', 'enter') else dict()
     on_dismiss: app.main_app.change_flow(id_of_flow('close', 'flow_popup'), **self.ae_closed_kwargs)
     auto_dismiss: True
@@ -173,7 +200,14 @@ WIDGETS = '''\
     ae_flow_id: id_of_flow('open', 'user_preferences')
     circle_fill_color: 0.69, 0.69, 0.99, 0.9
 
+
 <UserPreferencesOpenPopup@FlowDropDown>:
+    canvas.before:
+        Color:
+            rgba: (.69, .69, .69, 1.0)
+        RoundedRectangle:
+            pos: self.pos
+            size: self.size
     ChangeColorButton:
         color_name: 'flow_id_ink'
     ChangeColorButton:
@@ -190,6 +224,21 @@ WIDGETS = '''\
     # UserPrefSlider:    current kivy module vibrator.py does not support amplitudes arg of android api
     #    app_state_name: 'vibrate_amplitude'
     #    cursor_image: app.main_app.img_file('vibrate', app.ae_states['font_size'], app.ae_states['light_theme'])
+    BoxLayout:
+        size_hint_y: None
+        height: app.ae_states['font_size'] * 1.5 if installed_languages else 0
+        FlowButton:
+            ae_flow_id: id_of_flow('change', 'lang_code', self.text)
+            ae_clicked_kwargs: dict(popups_to_close=(self.parent.parent.parent, ))
+            square_fill_color: (.69, .69, .69, 1.0) if app.main_app.lang_code in ('', self.text) else Window.clearcolor
+            size_hint_x: 1
+            text: DEF_LANGUAGE
+        LangCodeButton:
+            lang_idx: 0
+        LangCodeButton:
+            lang_idx: 1
+        LangCodeButton:
+            lang_idx: 2
     BoxLayout:
         size_hint_y: None
         height: app.ae_states['font_size'] * 1.5
@@ -203,6 +252,17 @@ WIDGETS = '''\
             on_release: app.main_app.change_flow(id_of_flow('change', 'light_theme'), light_theme=True)
             color: THEME_LIGHT_FONT_COLOR or self.color
             square_fill_color: THEME_LIGHT_BACKGROUND_COLOR or self.square_fill_color
+    BoxLayout:
+        size_hint_y: None
+        height: app.ae_states['font_size'] * 1.5
+        DebugLevelButton:
+            level_idx: 0
+        DebugLevelButton:
+            level_idx: 1
+        DebugLevelButton:
+            level_idx: 2
+        DebugLevelButton:
+            level_idx: 3
 
 
 <UserPrefSlider@Slider>:
@@ -231,8 +291,8 @@ WIDGETS = '''\
     ae_clicked_kwargs: dict(popup_kwargs=dict(parent_popup_to_close=self.parent.parent, parent=self))
     square_fill_color: Window.clearcolor
 
-<FontSizeEditPopup@FlowDropDown>:
-    parent_popup_to_close: None
+
+<FontSizeEditPopup>:
     on_select:
         app.main_app.change_flow(id_of_flow('change', 'font_size'), \
         font_size=args[1], popups_to_close=(self.parent_popup_to_close, ))
@@ -251,6 +311,7 @@ WIDGETS = '''\
     FontSizeSelectButton:
         font_size: MAX_FONT_SIZE
 
+
 <FontSizeSelectButton@Button>:
     # text: f'Aa Bb Zz {round(self.font_size)}'      F-STRINGS don't work - displays always 15 as font size
     text: 'Aa Bb Zz {}'.format(round(self.font_size))
@@ -259,7 +320,7 @@ WIDGETS = '''\
     size: self.texture_size
     color: app.font_color
     background_normal: ''
-    background_color: Window.clearcolor
+    background_color: (.69, .69, .69, 1.0) if app.main_app.font_size == self.font_size else Window.clearcolor
 
 
 <ChangeColorButton@FlowButton>:
@@ -270,6 +331,7 @@ WIDGETS = '''\
     circle_fill_color: app.ae_states[self.color_name]
     text: self.color_name
 
+
 <ColorPickerOpenPopup@FlowDropDown>:
     ColorPicker:
         color: app.ae_states[root.attach_to.color_name] if root.attach_to else (0, 0, 0, 0)
@@ -282,10 +344,53 @@ WIDGETS = '''\
             RoundedRectangle:
                 pos: self.pos
                 size: self.size
+
+
+<LangCodeButton@OptionalButton>:
+    lang_idx: 0
+    ae_flow_id: id_of_flow('change', 'lang_code', self.text)
+    ae_clicked_kwargs: dict(popups_to_close=(self.parent.parent.parent, ))
+    square_fill_color: (.69, .69, .69, 1.0) if app.main_app.lang_code == self.text else Window.clearcolor
+    size_hint_x: 1 if self.visible else None
+    text: installed_languages[min(self.lang_idx, len(installed_languages) - 1)]
+    visible: len(installed_languages) > self.lang_idx
+
+
+<DebugLevelButton@FlowButton>:
+    level_idx: 0
+    ae_flow_id: id_of_flow('change', 'debug_level', self.text)
+    ae_clicked_kwargs: dict(popups_to_close=(self.parent.parent.parent, ))
+    square_fill_color: (.69, .69, .69, 1.0) if app.main_app.debug_level == self.level_idx else Window.clearcolor
+    size_hint_x: 1 if self.visible else None
+    text: DEBUG_LEVELS[min(self.level_idx, len(DEBUG_LEVELS) - 1)]
+    visible: len(DEBUG_LEVELS) > self.level_idx
 '''
 """ helper widgets with integrated app flow and observers ensuring change of app states (e.g. theme and size) """
 
 
+class FlowDropDown(DropDown):
+    """ drop down widget used for user selections from a list of items (represented by the children-widgets).
+
+    explicit class declaration for docs and for to allow initialization of ae_closed_kwargs attribute via __init__.
+    """
+
+    ae_closed_kwargs = DictProperty()   #: kwargs passed to all close action flow change event handlers
+
+
+class FlowPopup(Popup):
+    """ pop up widget used for dialogs and other top-most or modal windows.
+
+    explicit class declaration for docs and for to allow initialization of ae_closed_kwargs property via __init__.
+    """
+
+    ae_closed_kwargs = DictProperty()   #: kwargs passed to all close action flow change event handlers
+
+
+class FontSizeEditPopup(FlowDropDown):
+    """ drop down to select font size """
+    parent_popup_to_close = ObjectProperty()
+
+
 class FrameworkApp(App):
     """ kivy framework app class proxy redirecting events and callbacks to the main app class instance. """
 
@@ -340,10 +445,16 @@ class FrameworkApp(App):
         return self.main_app.call_method('on_key_release', keyboard.command_keys.get(key_code, str(key_code)))
 
     def on_start(self):
-        """ app start event """
-        self.win_pos_size_change()  # init. app./self.landscape (on app startup and after build)
+        """ kivy app start event.
+
+        Fired after call of :meth:`MainAppBase.run_app` method and MainAppBase.on_app_start event.
+
+        Kivy just created the main layout by calling its :meth:`~kivy.app.App.build` method and
+        attached it to the main window.
+        """
         self.main_app.framework_win = self.root.parent
-        self.main_app.call_method('on_app_start')
+        self.win_pos_size_change()  # init. app./self.landscape (on app startup and after build)
+        self.main_app.call_method('on_kivy_app_start')
 
     def on_pause(self) -> bool:
         """ app pause event """
@@ -358,9 +469,9 @@ class FrameworkApp(App):
         return True
 
     def on_stop(self):
-        """ quit app event """
+        """ quit app event (:meth:`MainAppBase.stop_app` emits the `on_app_stop` event) """
         self.main_app.save_app_states()
-        self.main_app.call_method('on_app_stop')
+        self.main_app.call_method('on_kivy_app_stop')
 
     def win_pos_size_change(self, *_):
         """ resize handler updates :attr:`~MainAppBase.win_rectangle` app state and :attr:`~FrameworkApp.landscape`. """
@@ -384,7 +495,6 @@ class KivyMainApp(MainAppBase):
         :return:                        callable for to start and stop/exit the GUI event loop.
         """
         Builder.load_string(WIDGETS)
-        self.call_method('on_app_init')
 
         win_rect = self.win_rectangle
         if win_rect:
@@ -427,7 +537,7 @@ class KivyMainApp(MainAppBase):
     def on_flow_widget_focused(self):
         """ set focus to the widget referenced by the current flow id. """
         liw = self.widget_by_flow_id(self.flow_id)
-        self.dpo(f"KivyMainApp.on_flow_widget_focused() '{self.flow_id}'"
+        self.vpo(f"KivyMainApp.on_flow_widget_focused() '{self.flow_id}'"
                  f" {liw} has={getattr(liw, 'focus', 'unsupported') if liw else ''}")
         if liw and getattr(liw, 'is_focusable', False) and not liw.focus:
             liw.focus = True
@@ -449,7 +559,7 @@ class KivyMainApp(MainAppBase):
 
     def play_sound(self, sound_name: str):
         """ play audio/sound file. """
-        self.dpo(f"KivyMainApp.play_sound {sound_name}")
+        self.vpo(f"KivyMainApp.play_sound {sound_name}")
         file: Optional[CachedFile] = self.find_sound(sound_name)
         if file:
             try:
@@ -465,7 +575,7 @@ class KivyMainApp(MainAppBase):
 
     def play_vibrate(self, pattern: Tuple = (0.03, 0.3)):
         """ play vibrate pattern. """
-        self.dpo(f"KivyMainApp.play_vibrate {pattern}")
+        self.vpo(f"KivyMainApp.play_vibrate {pattern}")
         try:        # added because is crashing with current plyer version (master should work)
             vibrator.pattern(pattern)
         # except jnius.jnius.JavaException as ex:
@@ -473,6 +583,15 @@ class KivyMainApp(MainAppBase):
         except Exception as ex:
             self.po(f"KivyMainApp.play_vibrate exception {ex}")
 
+    @staticmethod
+    def prevent_keyboard_covering(input_box_bottom: float):
+        """ prevent that the virtual keyboard popping up on mobile platforms is covering the text input field.
+
+        :param input_box_bottom:    y position of the bottom of the input field box.
+        """
+        keyboard_height = Window.keyboard_height or Window.height / 2  # 'or'-fallback because SDL2 reports 0 kbd height
+        Window.softinput_mode = 'below_target' if input_box_bottom < keyboard_height else ''
+
     def show_popup(self, popup_class: Type[Widget], **popup_attributes) -> Widget:
         """ open Popup using the `open` method. Overwriting the main app class method.
 
@@ -486,13 +605,12 @@ class KivyMainApp(MainAppBase):
         self.dpo(f"KivyAppBase.show_popup {popup_class} {popup_attributes}")
 
         parent = popup_attributes.pop('parent', self.framework_win)
-        popup_instance = popup_class()
-        for attr, value in popup_attributes.items():
-            setattr(popup_instance, attr, value)
+        popup_instance = popup_class(**popup_attributes)
 
         if not hasattr(popup_instance, 'close') and hasattr(popup_instance, 'dismiss'):
             popup_instance.close = popup_instance.dismiss   # create close() method alias for DropDown.dismiss() method
 
+        self.prevent_keyboard_covering(popup_instance.y)
         popup_instance.open(parent)
 
         return popup_instance
